__author__ = "Martin Uhrin"
__copyright__ = "Copyright 2022, Martin Uhrin"
__license__ = "LGPL-3.0-or-later"

version_info = (0, 1, 3)
__version__ = ".".join(map(str, version_info))
__maintainer__ = "Martin Uhrin"
__email__ = "martin.uhrin.10@ucl.ac.uk"
__status__ = "Development"

__all__ = "__author__", "__copyright__", "__version__", "__license__"



class _Primary(object):
    @property
    def mongos_mode(self):
        return "primary"

    @property
    def mode(self):
        return 0

    @property
    def name(self):
        return "Primary"

    @property
    def document(self):
        return {"mode": "primary"}

    @property
    def tag_sets(self):
        return [{}]

    @property
    def max_staleness(self):
        return -1

    @property
    def min_wire_version(self):
        return 0


def ensure_read_preference_type(key, value):
    """Raise a TypeError if the value is not a type compatible for ReadPreference."""
    for attr in ("document", "mode", "mongos_mode", "max_staleness"):
        if not hasattr(value, attr):
            raise TypeError(
                "{} must be an instance of {}".format(key, "pymongo.read_preference.ReadPreference")
            )


PRIMARY = _Primary()



import abc

from ._vendor.mongomock import store as mongomock_store


__all__ = ("ServerStore",)


class ServerStore(mongomock_store.ServerStore, metaclass=abc.ABCMeta):
    """MongoMock ServerStore with some additional functionality"""

    @abc.abstractmethod
    def close(self):
        """Close the store"""



from packaging import version
import warnings

from . import CollectionInvalid
from . import InvalidName
from . import OperationFailure
from .collection import Collection
from .filtering import filter_applies
from . import codec_options as mongomock_codec_options
from . import helpers
from . import read_preferences
from . import store

try:
    from pymongo import ReadPreference

    _READ_PREFERENCE_PRIMARY = ReadPreference.PRIMARY
except ImportError:
    _READ_PREFERENCE_PRIMARY = read_preferences.PRIMARY

try:
    from pymongo.read_concern import ReadConcern
except ImportError:
    from .read_concern import ReadConcern

_LIST_COLLECTION_FILTER_ALLOWED_OPERATORS = frozenset(["$regex", "$eq", "$ne"])


def _verify_list_collection_supported_op(keys):
    if set(keys) - _LIST_COLLECTION_FILTER_ALLOWED_OPERATORS:
        raise NotImplementedError(
            "list collection names filter operator {0} is not implemented yet in mongomock "
            "allowed operators are {1}".format(keys, _LIST_COLLECTION_FILTER_ALLOWED_OPERATORS)
        )


class Database(object):
    def __init__(
        self, client, name, _store, read_preference=None, codec_options=None, read_concern=None
    ):
        self.name = name
        self._client = client
        self._collection_accesses = {}
        self._store = _store or store.DatabaseStore()
        self._read_preference = read_preference or _READ_PREFERENCE_PRIMARY
        mongomock_codec_options.is_supported(codec_options)
        self._codec_options = codec_options or mongomock_codec_options.CodecOptions()
        if read_concern and not isinstance(read_concern, ReadConcern):
            raise TypeError("read_concern must be an instance of pymongo.read_concern.ReadConcern")
        self._read_concern = read_concern or ReadConcern()

    def __getitem__(self, coll_name):
        return self.get_collection(coll_name)

    def __getattr__(self, attr):
        if attr.startswith("_"):
            raise AttributeError(
                "%s has no attribute '%s'. To access the %s collection, use database['%s']."
                % (self.__class__.__name__, attr, attr, attr)
            )
        return self[attr]

    def __repr__(self):
        return "Database({0}, '{1}')".format(self._client, self.name)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._client == other._client and self.name == other.name
        return NotImplemented

    if helpers.PYMONGO_VERSION >= version.parse("3.12"):

        def __hash__(self):
            return hash((self._client, self.name))

    @property
    def client(self):
        return self._client

    @property
    def read_preference(self):
        return self._read_preference

    @property
    def codec_options(self):
        return self._codec_options

    @property
    def read_concern(self):
        return self._read_concern

    def _get_created_collections(self):
        return self._store.list_created_collection_names()

    if helpers.PYMONGO_VERSION < version.parse("4.0"):

        def collection_names(self, include_system_collections=True, session=None):
            warnings.warn("collection_names is deprecated. Use list_collection_names instead.")
            if include_system_collections:
                return list(self._get_created_collections())
            return self.list_collection_names(session=session)

    def list_collections(self, filter=None, session=None, nameOnly=False):
        raise NotImplementedError(
            "list_collections is a valid method of Database but has not been implemented in "
            "mongomock yet."
        )

    def list_collection_names(self, filter=None, session=None):
        """filter: only name field type with eq,ne or regex operator

        session: not supported
        for supported operator please see _LIST_COLLECTION_FILTER_ALLOWED_OPERATORS
        """
        field_name = "name"

        if session:
            raise NotImplementedError("Mongomock does not handle sessions yet")

        if filter:
            if not filter.get("name"):
                raise NotImplementedError(
                    "list collection {0} might be valid but is not "
                    "implemented yet in mongomock".format(filter)
                )

            filter = (
                {field_name: {"$eq": filter.get(field_name)}}
                if isinstance(filter.get(field_name), str)
                else filter
            )

            _verify_list_collection_supported_op(filter.get(field_name).keys())

            return [
                name
                for name in list(self._store._collections)
                if filter_applies(filter, {field_name: name}) and not name.startswith("system.")
            ]

        return [name for name in self._get_created_collections() if not name.startswith("system.")]

    def get_collection(
        self, name, codec_options=None, read_preference=None, write_concern=None, read_concern=None
    ):
        if read_preference is not None:
            read_preferences.ensure_read_preference_type("read_preference", read_preference)
        mongomock_codec_options.is_supported(codec_options)
        try:
            return self._collection_accesses[name].with_options(
                codec_options=codec_options or self._codec_options,
                read_preference=read_preference or self.read_preference,
                read_concern=read_concern,
                write_concern=write_concern,
            )
        except KeyError:
            self._ensure_valid_collection_name(name)
            collection = self._collection_accesses[name] = Collection(
                self,
                name=name,
                read_concern=read_concern,
                write_concern=write_concern,
                read_preference=read_preference or self.read_preference,
                codec_options=codec_options or self._codec_options,
                _db_store=self._store,
            )
            return collection

    def drop_collection(self, name_or_collection, session=None):
        if session:
            raise NotImplementedError("Mongomock does not handle sessions yet")
        if isinstance(name_or_collection, Collection):
            name_or_collection._store.drop()
        else:
            self._store[name_or_collection].drop()

    def _ensure_valid_collection_name(self, name):
        # These are the same checks that are done in pymongo.
        if not isinstance(name, str):
            raise TypeError("name must be an instance of str")
        if not name or ".." in name:
            raise InvalidName("collection names cannot be empty")
        if name[0] == "." or name[-1] == ".":
            raise InvalidName("collection names must not start or end with '.'")
        if "$" in name:
            raise InvalidName("collection names must not contain '$'")
        if "\x00" in name:
            raise InvalidName("collection names must not contain the null character")

    def create_collection(self, name, **kwargs):
        self._ensure_valid_collection_name(name)
        if name in self.list_collection_names():
            raise CollectionInvalid("collection %s already exists" % name)

        if kwargs:
            raise NotImplementedError("Special options not supported")

        self._store.create_collection(name)
        return self[name]

    def rename_collection(self, name, new_name, dropTarget=False):
        """Changes the name of an existing collection."""
        self._ensure_valid_collection_name(new_name)

        # Reference for server implementation:
        # https://docs.mongodb.com/manual/reference/command/renameCollection/
        if not self._store[name].is_created:
            raise OperationFailure('The collection "{0}" does not exist.'.format(name), 10026)
        if new_name in self._store:
            if dropTarget:
                self.drop_collection(new_name)
            else:
                raise OperationFailure(
                    'The target collection "{0}" already exists'.format(new_name), 10027
                )
        self._store.rename(name, new_name)
        return {"ok": 1}

    def dereference(self, dbref, session=None):
        if session:
            raise NotImplementedError("Mongomock does not handle sessions yet")

        if not hasattr(dbref, "collection") or not hasattr(dbref, "id"):
            raise TypeError("cannot dereference a %s" % type(dbref))
        if dbref.database is not None and dbref.database != self.name:
            raise ValueError(
                "trying to dereference a DBRef that points to "
                "another database (%r not %r)" % (dbref.database, self.name)
            )
        return self[dbref.collection].find_one({"_id": dbref.id})

    def command(self, command, **unused_kwargs):
        if isinstance(command, str):
            command = {command: 1}
        if "ping" in command:
            return {"ok": 1.0}
        # TODO(pascal): Differentiate NotImplementedError for valid commands
        # and OperationFailure if the command is not valid.
        raise NotImplementedError(
            "command is a valid Database method but is not implemented in Mongomock yet"
        )

    def with_options(
        self, codec_options=None, read_preference=None, write_concern=None, read_concern=None
    ):

        mongomock_codec_options.is_supported(codec_options)

        if write_concern:
            raise NotImplementedError(
                "write_concern is a valid parameter for with_options but is not implemented yet in "
                "mongomock"
            )

        if read_preference is None or read_preference == self._read_preference:
            return self

        return Database(
            self._client,
            self.name,
            self._store,
            read_preference=read_preference or self._read_preference,
            codec_options=codec_options or self._codec_options,
            read_concern=read_concern or self._read_concern,
        )



"""Tools for specifying BSON codec options."""

import collections
from packaging import version

from . import helpers

try:
    from bson import codec_options
except ImportError:
    codec_options = None


class TypeRegistry(object):
    pass


_FIELDS = (
    "document_class",
    "tz_aware",
    "uuid_representation",
    "unicode_decode_error_handler",
    "tzinfo",
)

if codec_options and helpers.PYMONGO_VERSION >= version.parse("3.8"):
    _DEFAULT_TYPE_REGISTRY = codec_options.TypeRegistry()
    _FIELDS = _FIELDS + ("type_registry",)
else:
    _DEFAULT_TYPE_REGISTRY = TypeRegistry()

# New default in Pymongo v4:
# https://pymongo.readthedocs.io/en/stable/examples/uuid.html#unspecified
if helpers.PYMONGO_VERSION >= version.parse("4.0"):
    _DEFAULT_UUID_REPRESENTATION = 0
else:
    _DEFAULT_UUID_REPRESENTATION = 3


class CodecOptions(collections.namedtuple("CodecOptions", _FIELDS)):
    def __new__(
        cls,
        document_class=dict,
        tz_aware=False,
        uuid_representation=None,
        unicode_decode_error_handler="strict",
        tzinfo=None,
        type_registry=None,
    ):

        if document_class != dict:
            raise NotImplementedError(
                "Mongomock does not implement custom document_class yet: %r" % document_class
            )

        if not isinstance(tz_aware, bool):
            raise TypeError("tz_aware must be True or False")

        if uuid_representation is None:
            uuid_representation = _DEFAULT_UUID_REPRESENTATION
        if uuid_representation != _DEFAULT_UUID_REPRESENTATION:
            raise NotImplementedError("Mongomock does not handle custom uuid_representation yet")

        if unicode_decode_error_handler not in ("strict", None):
            raise NotImplementedError(
                "Mongomock does not handle custom unicode_decode_error_handler yet"
            )

        if tzinfo:
            raise NotImplementedError("Mongomock does not handle custom tzinfo yet")

        values = (
            document_class,
            tz_aware,
            uuid_representation,
            unicode_decode_error_handler,
            tzinfo,
        )

        if "type_registry" in _FIELDS:
            if not type_registry:
                type_registry = _DEFAULT_TYPE_REGISTRY
            elif not type_registry == _DEFAULT_TYPE_REGISTRY:
                raise NotImplementedError(
                    "Mongomock does not handle custom type_registry yet %r" % type_registry
                )
            values = values + (type_registry,)

        return tuple.__new__(cls, values)

    def with_options(self, **kwargs):
        opts = self._asdict()
        opts.update(kwargs)
        return CodecOptions(**opts)


def is_supported(custom_codec_options):

    if not custom_codec_options:
        return None

    return CodecOptions(**custom_codec_options._asdict())



from urllib import parse
from typing import Union

from ._vendor import mongomock

from . import stores

__all__ = "MongoClient", "StoreType", "connect"


class StoreType:
    H5PY = "h5py"
    PYTABLES = "pytables"
    SQLITE = "sqlite"
    MEMORY = "memory"


def create_store(spec: Union[stores.ServerStore, str]) -> stores.ServerStore:
    """Server store factory"""
    if isinstance(spec, stores.ServerStore):
        return spec

    if isinstance(spec, str):
        uri = parse.urlparse(spec)
        store_type = StoreType.SQLITE  # The default

        if uri.query:
            qs_parsed = parse.parse_qs(uri.query)
            store_type = qs_parsed.pop("engine", store_type)

        if store_type == StoreType.H5PY:
            from . import h5py_store

            return h5py_store.ServerStore(uri.path)
        if store_type == StoreType.PYTABLES:
            from . import pytables_store

            return pytables_store.ServerStore(uri.path)
        if store_type == StoreType.SQLITE:
            from . import sqlite_store

            return sqlite_store.ServerStore(uri.path)
        if store_type == StoreType.MEMORY:
            from ._vendor.mongomock import store as mongomock_store

            return mongomock_store.ServerStore()

    raise TypeError(f"Unexpected server store specification: {spec}")


class MongoClient(mongomock.MongoClient):
    def __init__(
        self,
        store: stores.ServerStore,
        database: str = "",
        document_class=dict,
        tz_aware=False,
        read_preference=None,
    ):
        # Create a URI string with the (optional) default database
        uri = f"mongodb://localhost/{database}"
        super().__init__(
            host=uri,
            document_class=document_class,
            tz_aware=tz_aware,
            read_preference=read_preference,
            _store=create_store(store),
        )

    def __enter__(self) -> "MongoClient":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        super().close()
        # Make sure to close the store
        self._store.close()


def connect(
    path: str,
    document_class=dict,
    tz_aware=False,
    read_preference=None,
) -> MongoClient:
    """Convenience function to build a MongoClient.  This will create a server store based on the
    passed path string"""
    parsed = parse.urlparse(path)
    database = parsed.fragment
    return MongoClient(
        create_store(path),
        database=database,
        document_class=document_class,
        tz_aware=tz_aware,
        read_preference=read_preference,
    )
